#include "const.h"
#include "gen.h"
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>

struct args {
  Vertex *cube;
  Vertex *text;
  VertexI pos;
  // int is_text;
  int a;
  int i;
  int a3;
};

void gen_face(Vertex *cube, int i, int x, int y, int z, int x2, int y2, int z2,
              int *j, int is_text, Chunk dat, int ci, int a3) {
  VertexI a = {{x, y, z}};
  Vertex a2 = {{x, y, z}};

  if (is_text) {
    a.pos[i]--;
    int exists = 0;
    exists = (cube_exists(x, y, z, dat, ci) !=
              cube_exists(a.pos[0], a.pos[1], a.pos[2], dat, ci));

    a2.pos[0] = exists ? TEXT_GAP_H * i : -1.0;
    a2.pos[1] = exists ? TEXT_GAP_V * i : -1.0;
    a2.pos[2] = 0;
    if (!exists) {
      *j = (*j - 6 * a3);
      return;
    }
  }

  Vertex b = a2;
  Vertex c = a2;
  Vertex d = a2;

  if (is_text) {
    b.pos[0] += TEXT_GAP_H;
    c.pos[1] += TEXT_GAP_V;
    d.pos[0] += TEXT_GAP_H;
    d.pos[1] += TEXT_GAP_V;
  } else {
    b.pos[(i == 0) ? 1 : 0]++;
    c.pos[(i == 2) ? 1 : 2]++;
    d.pos[(i == 0) ? 1 : 0]++;
    d.pos[(i == 2) ? 1 : 2]++;
  }

  cube[*j] = a2;
  cube[*j + 1] = b;
  cube[*j + 2] = c;
  cube[*j + 3] = c;
  cube[*j + 4] = b;
  cube[*j + 5] = d;
}

void *gen_chunk(void *args) {

  struct args args2 = ((struct args *)args)[0];

  int i = args2.i;
  int a3 = args2.a3;
  int a2 = args2.a;
  int a = a2 % a3;

  int x2 = args2.pos.pos[0];
  int y2 = args2.pos.pos[1];
  int z2 = args2.pos.pos[2];
  Vertex *cube = args2.cube;
  Vertex *text = args2.text;
  // int is_text = args2.is_text;

  Chunk chunk =
      fetch_chunk(x2 / CHUNK_LENGTH, y2 / CHUNK_LENGTH, z2 / CHUNK_LENGTH, a);

  printf("Loading world... (%i / %i)\n", a2, CHUNK_ALL);

  if (i > CBLOCK_ALL * 18)
    return (void *)i;
  for (int y4 = 0; y4 < CHUNK_LENGTH; y4++) {

    for (int x4 = 0; x4 < CHUNK_LENGTH; x4++) {

      for (int z4 = 0; z4 < CHUNK_LENGTH; z4++) {
        int y3 = y2 + y4;
        int x3 = x2 + x4;
        int z3 = z2 + z4;

        for (int m = 0; m < 3; m++) {
          gen_face(cube, m, x3, y3, z3, x4, y4, z4, &i, 0, chunk, a, a3);
          gen_face(text, m, x3, y3, z3, x4, y4, z4, &i, 1, chunk, a, a3);
          i += 6 * a3;
          if (i > CBLOCK_ALL * 18)
            return (void *)i;
        }
      }
    }
  }
  return (void *)i;
}

int gen_cubes(Vertex *cube, Vertex *text, int x, int y, int z, int r, int is_purge) {
  Vertex noth = {{-1, -1, -1}};
  for (int i = 0; i < CBLOCK_ALL * 18; i++) {
    cube[i] = noth;
    text[i] = noth;
  }

  int i = 0;
  x *= CHUNK_LENGTH;
  y *= CHUNK_LENGTH;
  z *= CHUNK_LENGTH;
  int a3 = CHUNK_DIAMETER_H + (r * 2);
  int a = 0; //(z % CHUNK_DIAMETER_H) + (x % CHUNK_DIAMETER_H) * CHUNK_DIAMETER_H  + (y % CHUNK_DIAMETER_V) * CHUNK_DIAMETER_H * CHUNK_DIAMETER_H;
  //a = a % a3;
  //if (a < 0) a = -a;

  struct args args[a3];
  int thread_return[a3];

  pthread_t thread_id[a3];
  for (int y2 = y - (CHUNK_RADIUS_V + r) * CHUNK_LENGTH;
       y2 <= y + (CHUNK_RADIUS_V + r) * CHUNK_LENGTH; y2 += CHUNK_LENGTH) {

    for (int x2 = x - (CHUNK_RADIUS_H + r) * CHUNK_LENGTH;
         x2 <= x + (CHUNK_RADIUS_H + r) * CHUNK_LENGTH; x2 += CHUNK_LENGTH) {

      for (int z2 = z - (CHUNK_RADIUS_H + r) * CHUNK_LENGTH;
           z2 <= z + (CHUNK_RADIUS_H + r) * CHUNK_LENGTH; z2 += CHUNK_LENGTH) {

        VertexI pos = {{x2, y2, z2}};
        int a2 = a % a3;
        args[a2].cube = cube;
        args[a2].text = text;
        args[a2].pos = pos;
        // args[a2].is_text = is_text;
        args[a2].a = a;
        args[a2].i = i;
        args[a2].a3 = a3;

        /*gen_chunk((void *)&args); */

        if (is_purge) purge_chunks(a2);
        pthread_create(&(thread_id[a2]), NULL, gen_chunk, (void *)&(args[a2]));

        i += 6;

        a++;
      }

      for (int k = 0; k < a3; k++) {
        pthread_join((thread_id[k]), (void **)&thread_return[k]);
      }

      int max = 0;
      int min = 0xFFFFFFF;

      for (int a = 0; a < a3; a++) {
        max = (max > thread_return[a]) ? max : thread_return[a];
        min = (min < thread_return[a]) ? min : thread_return[a];
      }
      i = max + 6;
    }
  }

  /*  if (i < CBLOCK_ALL * 9 && r < MAX_R) {
      printf("Expanding radius...\n");
      return gen_cubes(cube, text, x / CHUNK_LENGTH, y / CHUNK_LENGTH,
                       z / CHUNK_LENGTH, r + 1);
    }
  */

  return i;
}